from hashlib import sha256
from getpass import getpass
from sys import exit
from tkinter import Tk, filedialog
from argparse import ArgumentParser
from numpy.random import RandomState, Generator
from os import mkdir, remove, removedirs, sep
from tarfile import open as tarOpen, ReadError
from os.path import dirname, basename, abspath, exists, isfile
from logging import Formatter, StreamHandler, getLogger, DEBUG, INFO, CRITICAL

def Poodle (path, rng):
    with open(path, "rb") as file:
        poodleBinary = file.read()
    with open(path, "wb") as file:
        file.write(bytes(a ^ b for a, b in zip(poodleBinary, rng.bytes(len(poodleBinary)))))

def main ():
    parser = ArgumentParser(prog="python poodle.py", description="A Custom Encryption Program")
    parser.add_argument("-v", "--debug", help="Enter debug mode", action="store_true")
    parser.add_argument("-q", "--quiet", help="Suppress all output except for errors", action="store_true")
    parser.add_argument("-s", "--silent", help="Suppress all output including errors", action="store_true")
    parser.add_argument("-r", "--recursive", help="Recursively encrypt the data in a given directory", action="store_true")
    parser.add_argument("-d", "--delete", help="Delete input file(s) or directory", action="store_true")
    parser.add_argument("-x", "--extract", help="Decrypt a poodle file", action="store_true")
    parser.add_argument("-m", "--mode", help="Specify the rng used by poodle. 1: MT19337, uses a 32bit integer as the seed; 2: PCG64, uses a 64bit integer as the seed (more secure). Default is 1.", default=1, const=1, type=int, nargs="?")
    parser.add_argument("-p", "--password", help="Specify password used to poodle. Note, this use of this option is insecure. Only use if you are confident your command history is confidential.", type=str)
    parser.add_argument("-S", "--seed", help="Specify the hex seed used to poodle. Include leading zeros. If mode is 1, use a 32bit seed. If 2, a 64bit seed. Note, this use of this option is insecure. Only use if you are confident your command history is confidential.", type=str)
    parser.add_argument("-i", "--input", help="Specify name of file(s) or directory to poodle", type=str, nargs="+")
    parser.add_argument("-o", "--output", help="Specify name to give the output poodle or the directory to extract the poodle to", type=str)
    args = parser.parse_args()

    formatter = Formatter("%(levelname)s %(funcName)s(%(lineno)d) %(message)s")
    consoleHandler = StreamHandler()
    consoleHandler.setFormatter(formatter)
    logger = getLogger("root")
    if args.debug: logger.setLevel(DEBUG)
    elif args.quiet: logger.setLevel(CRITICAL)
    elif args.silent: logger.disabled = True
    else: logger.setLevel(INFO)
    logger.addHandler(consoleHandler)

    # directory = dirname(executable)
    directory = dirname(__file__)
    
    logger.debug(f"Running in {directory}")
    logger.debug(f"Arguments: {args}")

    if args.input == None or len(args.input) == 0:
        logger.debug("Creating selection dialog")
        root = Tk()
        root.withdraw()
        root.attributes("-topmost", True)
        if args.recursive and not args.extract:
            logger.info("Choose a Directory to Poodle")
            args.input = [filedialog.askdirectory(title="Choose a Directory to Poodle", initialdir=directory)]
        elif not args.extract:
            args.input = []
            logger.info("Choose Files to Poodle")
            for file in filedialog.askopenfiles(title="Choose Files to Poodle", initialdir=directory):
                args.input.append(file.name)
                file.close()
        else:
            try:
                logger.info("Choose a Poodle File to Decrypt")
                file = filedialog.askopenfile(title="Choose a Poodle File to Decrypt", filetypes=[("All files", "*.*")], initialdir=directory)
                args.input = [file.name]
                file.close()
            except AttributeError:
                logger.critical("No poodle file selected")
                exit(0)
        root.destroy()
    elif len(args.input) > 1 and args.extract:
        logger.critical("Only one poodle at a time")
        exit(1)

    logger.debug(f"Input paths: {args.input}")

    if len(args.input) == 0:
        logger.critical("No file(s) or directory to poodle")
        exit(0)

    for path in args.input:
        if not exists(path):
            logger.critical(f"The path {path} does not exist")
            exit(1)
        elif (not isfile(path) and __file__.startswith(abspath(path))) or (isfile(path) and abspath(path) == __file__):
            logger.critical("Cannot poodle poodle.exe")
            exit(1)

    if args.seed == None:
        try:
            while not args.password:
                args.password = getpass()
                if not args.extract:
                    if getpass("Confirm Password: ") != args.password:
                        logger.critical("Passwords don't match. Try again. Use Ctrl+C to quit.")
                        args.password = None
                        continue
                break
        except KeyboardInterrupt:
            print("")
            logger.critical("Aborting")
            exit(0)
    
        m = sha256()
        m.update(args.password.encode("utf-16"))
        logger.debug(f"The password is {args.password}")
        args.seed = 0
        #? The seed for the one-time-pad is generated by xor'ing each byte in the digest by its corresponding byte in the seed
        match args.mode:
            case 1: #? MT19337
                for index, byte in enumerate(m.digest()): args.seed ^= byte << ((index % 4) << 3)
            case 2: #? PCG64
                for index, byte in enumerate(m.digest()): args.seed ^= byte << ((index % 8) << 3)
            case _:
                logger.critical(f"Mode can only be 1 or 2 not {args.mode}")
                exit(1)
    elif (args.mode == 1 and (len(args.seed) == 8 or (len(args.seed) == 10 and args.seed[:2] == "0x"))) or (args.mode == 2 and (len(args.seed) == 16 or (len(args.seed) == 18 and args.seed[:2] == "0x"))):
        try:
            args.seed = int(args.seed, 16)
        except ValueError:
            logger.critical(f"Could not parse the given seed {args.seed}")
            exit(1)
    match args.mode:
        case 1:
            rng = RandomState(args.seed)
        case 2:
            rng = Generator(args.seed)
    logger.debug(f"The seed is 0x{hex(args.seed)[2:].zfill(8)}")
    
    if args.output == None and not args.extract:
        logger.debug("Creating save dialog")
        root = Tk()
        root.withdraw()
        root.attributes("-topmost", True)
        try:
            logger.info("Save the Poodle File")
            with filedialog.asksaveasfile(title="Save the Poodle File", initialfile="poodle", filetypes=[("All files", "*.*")], initialdir=dirname(abspath(args.input[-1]))) as tar:
                args.output = tar.name
        except AttributeError:
            logger.critical("No poodle saved")
            exit(0)
        root.destroy()
        if args.output == None:
            logger.critical("No file to save poodle")
            exit(0)
    elif args.output == None and args.extract: args.output = dirname(args.input[0])
    elif args.output != None and (args.extract or args.output.endswith(sep)):
        logger.debug("Creating output directory")
        try:
            mkdir(args.output)
        except PermissionError:
            logger.critical(f"poodle.exe did not have the correct permissions to create the output directory {args.output}")
            exit(1)
        except FileNotFoundError:
            logger.critical(f"A part of the output path, {args.output}, does not exist")
            exit(1)
        except FileExistsError:
            pass

    logger.debug(f"Output path: {args.output}")

    hadReadError = False
    if not args.extract:
        try:
            logger.debug(f"Creating poodle at {args.output}")
            with tarOpen(args.output, "w:gz") as tar:
                for path in args.input:
                    logger.debug(f"Adding {path}")
                    tar.add(path, arcname=basename(path))
            Poodle(args.output, rng)
        except PermissionError:
            logger.critical(f"poodle.exe did not have the correct permissions to create the poodle file {args.output}")
            exit(1)
        except OSError as e:
            logger.critical(e)
            logger.info(f"Removing the incomplete poodle file at {args.output}")
            remove(args.output)
            exit(1)
    else:
        try:
            logger.debug(f"Extracting poodle at {args.input[0]}")
            Poodle(args.input[0], rng)
            with tarOpen(args.input[0], "r:gz") as tar:
                tar.extractall(args.output)
        except ReadError:
            logger.critical(f"ReadError occurred when trying to read {args.input[0]}. It is likely the provided password was incorrect.")
            hadReadError = True
        except PermissionError:
            logger.critical(f"poodle.exe did not have the correct permissions to read the poodle file {args.input[0]}")
            hadReadError = True
        except OSError as e:
            logger.critical(e)
            hadReadError = True
        except Exception as e:
            logger.critical(e)
            hadReadError = True
        try:
            logger.debug("Reverting the poodle file to its original state")
            rng.seed(args.seed)
            Poodle(args.input[0], rng)
        except PermissionError:
            logger.critical(f"poodle.exe did not have the correct permissions to read the poodle file {args.input[0]}")
        except OSError as e:
            logger.critical(e)

    if args.delete and not hadReadError:
        logger.info("Deleting input file(s)/directory")
        for path in args.input:
            logger.debug(f"Deleting {path}")
            if isfile(path): remove(path)
            else: removedirs(path)
    elif args.delete:
        logger.info("Not deleting the input poodle because there was a read error")
        exit(1)
    elif hadReadError: exit(1)
    
    logger.info("Success")

if __name__ == "__main__":
    main()
